@@ -6,17 +6,14 @@ from contextlib import contextmanager, ExitStack
 from copy import deepcopy
 from pathlib import Path
 from tempfile import NamedTemporaryFile
-from typing import Dict, TextIO
+from typing import Dict, Optional, TextIO
 
 from IPython.core.interactiveshell import InteractiveShell
 from nbformat import read
 
 from autograde.helpers import import_filter
 from autograde.static import INJECT_BEFORE, INJECT_AFTER
-from autograde.util import logger, capture_output, timeout, StopWatch
-
-
-# Globals and constants variables.
+from autograde.util import logger, capture_output, deadline, StopWatch
 
 
 def as_py_comment(s: str, indentation: int = 0):
@@ -35,6 +32,7 @@ def as_py_comment(s: str, indentation: int = 0):
 
 class ArtifactLoader:
     """Helper class that provides a dict like interface for accessing artifact files"""
+
     def __init__(self, root='artifacts'):
         self._root = Path(root)
 
@@ -44,7 +42,7 @@ class ArtifactLoader:
 
 
 @contextmanager
-def shadowed_exec(source: str, *args, **kwargs):
+def shadow_exec(source: str, *args, **kwargs):
     """
     Wrapper for builtin `exec` that accepts source code as input and executes it while preserving
     a copy of it in the file system. That's particularly useful when inspecting the state created
@@ -56,7 +54,7 @@ def shadowed_exec(source: str, *args, **kwargs):
     :return: path of the source file
     """
     source = f'{source}\n'
-    with NamedTemporaryFile(mode='wt') as shadow_copy:
+    with NamedTemporaryFile(mode='wt', encoding='utf-8') as shadow_copy:
         shadow_copy.write(source)
         shadow_copy.flush()
 
@@ -68,7 +66,7 @@ def shadowed_exec(source: str, *args, **kwargs):
 
 @contextmanager
 def exec_notebook(notebook, file: TextIO = sys.stdout, cell_timeout: float = 0.,
-                  ignore_errors: bool = False, variables: Dict = None):
+                  ignore_errors: bool = False, variables: Optional[Dict] = None):
     """
     Extract source code from jupyter notebook and execute it.
 
@@ -104,7 +102,7 @@ def exec_notebook(notebook, file: TextIO = sys.stdout, cell_timeout: float = 0.,
                 # render code
                 source = shell.input_transformer_manager.transform_cell(cell.source)
                 yield (
-                    f'nb-{i+1}',
+                    f'nb-{i + 1}',
                     f'{source.strip()}\n\n# injected by test\ndump_figure()',
                     cell_timeout
                 )
@@ -125,56 +123,50 @@ def exec_notebook(notebook, file: TextIO = sys.stdout, cell_timeout: float = 0.,
 
     # actual code execution
     with ExitStack() as shadow_stack:
-        for i, (label, code, timeout_) in enumerate(code_cells, start=1):
+        for i, (label, code_cell, timeout) in enumerate(code_cells, start=1):
             state.update({'__LABEL__': deepcopy(label), '__PLOT_REGISTRY__': []})
 
             with io.StringIO() as stdout, io.StringIO() as stderr:
                 logger.debug(f'[{i}/{len(code_cells)}] execute cell ("{label}")')
                 stopwatch = StopWatch()
 
+                # state transmuting code execution
                 try:
-                    with capture_output(stdout, stderr):
-                        # actual execution that extends state
-                        with ExitStack() as es:
-                            if if_regex is not None and i > 1:
-                                es.enter_context(import_filter(if_regex, blacklist=if_blacklist))
-                            es.enter_context(timeout(timeout_))
-                            es.enter_context(stopwatch)
-
-                            shadow_stack.enter_context(shadowed_exec(code, state))
+                    with capture_output(stdout, stderr), ExitStack() as execution_stack:
+                        if if_regex is not None and i > 1:
+                            execution_stack.enter_context(import_filter(if_regex, blacklist=if_blacklist))
+                        execution_stack.enter_context(deadline(timeout))
+                        execution_stack.enter_context(stopwatch)
+                        shadow_stack.enter_context(shadow_exec(code_cell, state))
 
+                # extend log with a meaningful error message
                 except Exception as error:
-                    # extend log with some meaningful error message
                     traceback.print_exception(type(error), error, error.__traceback__, file=stderr)
 
                     if not ignore_errors:
+                        print('abort due to previous error', file=stderr)
                         raise error
 
+                # log code cell and outputs
                 finally:
-                    # log code and output
                     with capture_output(file):
                         _label = f' CODE CELL {label} '
                         print(f'# {_label:-^78}')
-                        print(str(code).strip())
+                        print(f'{str(code_cell).strip()}\n')
+                        print(f"# EXECUTED IN {stopwatch.duration_rel()[-1]:.3}s")
 
-                        print(f"\n# EXECUTED IN {stopwatch.duration_rel()[-1]:.3}s")
-
-                        stdout_s = stdout.getvalue()
-                        if stdout_s:
+                        if stdout_s := stdout.getvalue():
                             print('# STDOUT')
                             print(as_py_comment(stdout_s, 4))
 
-                        stderr_s = stderr.getvalue()
-                        if stderr_s:
+                        if stderr_s := stderr.getvalue():
                             print('# STDERR')
                             print(as_py_comment(stderr_s, 4))
 
-                        print('\n')
+                        print()
 
-        # add markdown comments to state
+        # add special items to state
         state['__COMMENTS__'] = md_cells
-
-        # add artifact loader
         state['__ARTIFACTS__'] = ArtifactLoader()
 
         logger.debug('execution completed')
