@@ -1,20 +1,32 @@
-import datetime
+import base64
 import logging
 import math
 import os
 import re
 import sys
-import tarfile
 import time
 from contextlib import contextmanager
+from datetime import datetime
 from pathlib import Path
 from tempfile import TemporaryDirectory
-from typing import Iterable, Union, List
+from typing import ContextManager, Generator, Iterable, Union, List
+from zipfile import ZipFile
 
-import pytz
+from htmlmin.minify import html_minify
+from jinja2 import Environment, PackageLoader, select_autoescape
 
+import autograde
+from autograde.static import CSS, FAVICON
+
+DateTime = datetime
 ALPHA_NUMERIC = re.compile(r'[^\w]')
-
+FAVICON = base64.b64encode(FAVICON).decode('utf-8')
+JINJA_ENV = Environment(
+    loader=PackageLoader('autograde', 'templates'),
+    autoescape=select_autoescape(['html', 'xml']),
+    trim_blocks=True,
+    lstrip_blocks=True
+)
 
 _formatter = logging.Formatter(
     '{asctime} [{levelname}] {processName}:  {message}',
@@ -38,8 +50,8 @@ def parse_bool(s):
     raise ValueError(f'cannot parse "{s}" as boolean')
 
 
-def timestamp_utc_iso():
-    return datetime.datetime.now(pytz.utc).replace(microsecond=0).isoformat()
+def now() -> DateTime:
+    return DateTime.now().replace(microsecond=0)
 
 
 def loglevel(x):
@@ -51,6 +63,10 @@ def project_root():
     return Path(autograde.__file__).parent.parent
 
 
+def float_equal(a, b):
+    return math.isclose(a, b) or (math.isnan(a) and math.isnan(b))
+
+
 def _alpha_numeric_split(s):
     for w in ALPHA_NUMERIC.split(s.strip()):
         if w:
@@ -103,8 +119,10 @@ def capture_output(tmp_stdout=None, tmp_stderr=None):
 
 
 @contextmanager
-def cd(path, mkdir=False):
-    cwd = os.getcwd()
+def cd(path: Union[Path, str], mkdir: bool = False) -> ContextManager[Path]:
+    """Change directory"""
+    path = Path(path)
+    cwd = Path(os.getcwd())
 
     if mkdir:
         logger.debug(f'create directories: {path}')
@@ -122,38 +140,29 @@ def cd(path, mkdir=False):
 
 
 @contextmanager
-def mount_tar(path, mode='r'):
-    prefix = mode[0]
+def cd_zip(path: Union[Path, str], mode: str = 'w') -> ContextManager[ZipFile]:
+    """Change to a temporary directory and write all files to zip file when context is left"""
+    if mode not in set('wax'):
+        raise ValueError('supported modes: "w", "a", "x"')
 
-    # TODO use ValueError instead
-    assert prefix in ['r', 'w', 'a'], f'unknown prefix {prefix}'
-    assert '|' not in mode, 'streaming is not supported'
+    with ZipFile(Path(path), mode=mode) as zipf, TemporaryDirectory() as tmp, cd(tmp):
+        tmp = Path(tmp)
+        yield zipf
 
-    with TemporaryDirectory() as tempdir:
-        logger.debug(f'create mount point at {tempdir} in mode "{mode}"')
-        try:
-            if prefix in ['r', 'a']:
-                logger.debug(f'extract files from {path}')
-                with tarfile.open(path, mode='r'+mode[1:]) as tar:
-                    tar.extractall(tempdir)
-
-            yield tempdir
-
-        finally:
-            if prefix in ['w', 'a']:
-                logger.debug(f'write changes to {path}')
-                with tarfile.open(path, mode='w'+mode[1:]) as tar:
-                    tar.add(tempdir, arcname='')
+        # add all files in temporary directory to zipfile
+        for file in filter(Path.is_file, tmp.rglob('*')):
+            zipf.write(file, file.relative_to(tmp))
 
 
 class StopWatch:
     """Measure durations"""
+
     def __init__(self):
-        self._captures = [time.time()]
+        self._captures = [time.monotonic()]
 
     def capture(self) -> int:
         """Store current time and return index of capture"""
-        self._captures.append(time.time())
+        self._captures.append(time.monotonic())
         return len(self._captures) - 1
 
     def __enter__(self):
@@ -177,23 +186,75 @@ class StopWatch:
 
 
 @contextmanager
-def timeout(timeout_):
-    start = time.time()
+def deadline(timeout):
+    """Context that fails if not left in time"""
+    start = time.monotonic()
 
     # trace callbacks
     def _globaltrace(frame, event, arg):
         return _localtrace if event == 'call' else None
 
     def _localtrace(frame, event, arg):
-        if time.time() - start >= timeout_ and event == 'line':
-            raise TimeoutError(f'code execution took longer than {timeout_:.3f}s to terminate')
+        if time.monotonic() - start >= timeout and event == 'line':
+            logger.debug('abort execution due to timeout')
+            raise TimeoutError(f'code execution took longer than {timeout:.3f}s to terminate')
+
+    trace = sys.gettrace()
 
     # activate tracing only in case timeout was actually set
-    if timeout_:
+    if timeout:
         sys.settrace(_globaltrace)
 
     try:
         yield start
 
     finally:
-        sys.settrace(None)
+        sys.settrace(trace)
+
+
+class WatchDog:
+    """Observe directory for file changes without looking into files"""
+
+    @staticmethod
+    def _hash_stat(stat: os.stat_result) -> int:
+        return hash(stat.st_ctime) ^ hash(stat.st_mtime_ns) ^ hash(stat.st_size)
+
+    def __init__(self, path: Union[Path, str]):
+        self._path = Path(path)
+
+        if not self._path.is_dir():
+            raise ValueError(f'{self._path} is no valid directory')
+
+        self._index = dict()
+        self.reload()
+
+    def _list(self):
+        yield from ((p, self._hash_stat(p.stat())) for p in self._path.rglob('*') if p.is_file())
+
+    def reload(self):
+        """Re-build index (discards all changes that have been registered)"""
+        self._index = dict(self._list())
+
+    def list_changed(self) -> Generator[Path, None, None]:
+        """List files that have been modified since the index was built"""
+        yield from (path for path, hsh in self._list() if hsh != self._index.get(path))
+
+    def list_not_changed(self) -> Generator[Path, None, None]:
+        """List files that have NOT been modified since the index was built"""
+        yield from (path for path, hsh in self._list() if hsh == self._index.get(path))
+
+
+def render(template: str, minify: bool = True, **kwargs):
+    """Render template with default values set"""
+    html = JINJA_ENV.get_template(template).render(
+        autograde=autograde,
+        css=CSS,
+        favicon=FAVICON,
+        timestamp=now().isoformat(),
+        **kwargs
+    )
+
+    if minify:
+        return html_minify(html)
+
+    return html
